@@ -17,6 +17,12 @@ import os
 import json
 import base64
 import asyncio
+from functools import reduce
+
+# Not available in Python 3.7, and slixmpp already imports the right things
+# for me
+from slixmpp.types import TypedDict
+
 from slixmpp.plugins.xep_0060.stanza import Items, EventItems
 from slixmpp.plugins.xep_0004 import Form
 from slixmpp.plugins.base import BasePlugin, register_plugin
@@ -108,9 +114,10 @@ def _generate_encrypted_payload(encrypted) -> Encrypted:
 
     tag['header']['sid'] = str(encrypted['sid'])
     tag['header']['iv']['value'] = b64enc(encrypted['iv'])
-    tag['payload']['value'] = b64enc(encrypted['payload'])
+    if 'payload' in encrypted:
+        tag['payload']['value'] = b64enc(encrypted['payload'])
 
-    for bare_jid, devices in encrypted['keys'].items():
+    for devices in encrypted['keys'].values():
         for rid, device in devices.items():
             key = Key()
             key['value'] = b64enc(device['data'])
@@ -151,6 +158,9 @@ class MissingOwnKey(XEP0384): pass
 class NoAvailableSession(XEP0384): pass
 
 
+class UninitializedOMEMOSession(XEP0384): pass
+
+
 class EncryptionPrepareException(XEP0384):
     def __init__(self, errors):
         self.errors = errors
@@ -174,6 +184,11 @@ class ErroneousPayload(XEP0384):
     """To be raised when the payload is not of the form we expect"""
 
 
+class ErroneousParameter(XEP0384):
+    """To be raised when parameters to the `encrypt_message` method aren't
+    used as expected."""
+
+
 class XEP_0384(BasePlugin):
 
     """
@@ -182,12 +197,15 @@ class XEP_0384(BasePlugin):
 
     name = 'xep_0384'
     description = 'XEP-0384 OMEMO'
-    dependencies = {'xep_0004', 'xep_0060', 'xep_0163'}
+    dependencies = {'xep_0004', 'xep_0030', 'xep_0060', 'xep_0163', 'xep_0334'}
     default_config = {
         'data_dir': None,
         'storage_backend': None,
         'otpk_policy': DefaultOTPKPolicy,
         'omemo_backend': SignalBackend,
+        'auto_heartbeat': True,
+        'heartbeat_after': 53,
+        # TODO: 'drop_inactive_after': 300,
     }
 
     backend_loaded = HAS_OMEMO and HAS_OMEMO_BACKEND
@@ -195,6 +213,12 @@ class XEP_0384(BasePlugin):
     # OMEMO Bundles used for encryption
     bundles = {}  # type: Dict[str, Dict[int, ExtendedPublicBundle]]
 
+    # Used at startup to prevent publishing device list and bundles multiple times
+    _initial_publish_done = False
+
+    # Initiated once the OMEMO session is created.
+    __omemo_session: Optional[SessionManager] = None
+
     def plugin_init(self) -> None:
         if not self.backend_loaded:
             log_str = ("xep_0384 cannot be loaded as the backend omemo library "
@@ -211,30 +235,17 @@ class XEP_0384(BasePlugin):
             raise PluginCouldNotLoad("xep_0384 cannot be loaded as there is "
                                      "no data directory specified.")
 
-        storage = self.storage_backend
-        if self.storage_backend is None:
-            storage = JSONFileStorage(self.data_dir)
-
-        otpkpolicy = self.otpk_policy
-        bare_jid = self.xmpp.boundjid.bare
         self._device_id = _load_device_id(self.data_dir)
-
-        try:
-            self._omemo = SessionManager.create(
-                storage,
-                otpkpolicy,
-                self.omemo_backend,
-                bare_jid,
-                self._device_id,
-            )
-        except:
-            log.error("Couldn't load the OMEMO object; ¯\\_(ツ)_/¯")
-            raise PluginCouldNotLoad
+        asyncio.ensure_future(self.session_start_omemo())
 
         self.xmpp.add_event_handler('session_start', self.session_start)
         self.xmpp['xep_0060'].map_node_event(OMEMO_DEVICES_NS, 'omemo_device_list')
         self.xmpp.add_event_handler('omemo_device_list_publish', self._receive_device_list)
-        return None
+
+        # If this plugin is loaded after 'session_start' has fired, we still
+        # need to publish bundles
+        if self.xmpp.is_connected and not self._initial_publish_done:
+            asyncio.ensure_future(self._initial_publish())
 
     def plugin_end(self):
         if not self.backend_loaded:
@@ -244,19 +255,51 @@ class XEP_0384(BasePlugin):
         self.xmpp.remove_event_handler('omemo_device_list_publish', self._receive_device_list)
         self.xmpp['xep_0163'].remove_interest(OMEMO_DEVICES_NS)
 
+    async def session_start_omemo(self):
+        """Creates the OMEMO session object"""
+
+        storage = self.storage_backend
+        if self.storage_backend is None:
+            storage = JSONFileStorage(self.data_dir)
+
+        otpkpolicy = self.otpk_policy
+        bare_jid = self.xmpp.boundjid.bare
+
+        try:
+            self.__omemo_session = await SessionManager.create(
+                storage,
+                otpkpolicy,
+                self.omemo_backend,
+                bare_jid,
+                self._device_id,
+            )
+        except Exception as exn:
+            log.error("Couldn't load the OMEMO object; ¯\\_(ツ)_/¯")
+            raise PluginCouldNotLoad from exn
+
+    def _omemo(self) -> SessionManager:
+        """Helper method to unguard potentially uninitialized SessionManager"""
+        if self.__omemo_session is None:
+            raise UninitializedOMEMOSession
+        return self.__omemo_session
+
     async def session_start(self, _jid):
+        await self._initial_publish()
+
+    async def _initial_publish(self):
         if self.backend_loaded:
             self.xmpp['xep_0163'].add_interest(OMEMO_DEVICES_NS)
             await asyncio.wait([
                 self._set_device_list(),
                 self._publish_bundle(),
             ])
+            self._initial_publish_done = True
 
     def my_device_id(self) -> int:
         return self._device_id
 
-    def my_fingerprint(self) -> str:
-        bundle = self._omemo.public_bundle.serialize(self.omemo_backend)
+    async def my_fingerprint(self) -> str:
+        bundle = await self._omemo().public_bundle.serialize(self.omemo_backend)
         return fp_from_ik(bundle['ik'])
 
     def _set_node_config(
@@ -306,10 +349,10 @@ class XEP_0384(BasePlugin):
         )
 
     async def _generate_bundle_iq(self, publish_options: bool = True) -> Iq:
-        bundle = self._omemo.public_bundle.serialize(self.omemo_backend)
+        bundle = self._omemo().public_bundle.serialize(self.omemo_backend)
 
         jid = self.xmpp.boundjid
-        disco = await self.xmpp['xep_0030'].get_info(jid.bare)
+        disco = await self.xmpp['xep_0030'].get_info(jid=jid.bare, local=False)
         publish_options = PUBLISH_OPTIONS_NODE in disco['disco_info'].get_features()
 
         iq = self.xmpp.Iq(stype='set')
@@ -344,14 +387,16 @@ class XEP_0384(BasePlugin):
         return iq
 
     async def _publish_bundle(self) -> None:
-        if self._omemo.republish_bundle:
+        log.debug('Publishing our own bundle. Do we need to?')
+        if self._omemo().republish_bundle:
+            log.debug('Publishing.')
             iq = await self._generate_bundle_iq()
             try:
                 await iq.send()
-            except IqError as e:
+            except IqError as exn:
                 # TODO: Slixmpp should handle pubsub#errors so we don't have to
                 # fish the element ourselves
-                precondition = e.iq['error'].xml.find(
+                precondition = exn.iq['error'].xml.find(
                     '{%s}%s' % (PUBSUB_ERRORS, 'precondition-not-met'),
                 )
                 if precondition is not None:
@@ -367,8 +412,11 @@ class XEP_0384(BasePlugin):
                         raise
                     iq = await self._generate_bundle_iq(publish_options=False)
                     await iq.send()
+            else:
+                log.debug('Not publishing.')
 
     async def _fetch_bundle(self, jid: str, device_id: int) -> Optional[ExtendedPublicBundle]:
+        log.debug('Fetching bundle for JID: %r, device: %r', jid, device_id)
         node = '%s:%d' % (OMEMO_BUNDLES_NS, device_id)
         try:
             iq = await self.xmpp['xep_0060'].get_items(jid, node)
@@ -380,16 +428,17 @@ class XEP_0384(BasePlugin):
 
     async def _fetch_device_list(self, jid: JID) -> None:
         """Manually query PEP OMEMO_DEVICES_NS nodes"""
+        log.debug('Fetching device list for JID: %r', jid)
         iq = await self.xmpp['xep_0060'].get_items(jid.full, OMEMO_DEVICES_NS)
         return await self._read_device_list(jid, iq['pubsub']['items'])
 
-    def _store_device_ids(self, jid: str, items: Union[Items, EventItems]) -> None:
+    async def _store_device_ids(self, jid: str, items: Union[Items, EventItems]) -> None:
         """Store Device list"""
         device_ids = []  # type: List[int]
         items = list(items)
         if items:
             device_ids = [int(d['id']) for d in items[0]['devices']]
-        return self._omemo.newDeviceList(str(jid), device_ids)
+        return await self._omemo().newDeviceList(str(jid), device_ids)
 
     def _receive_device_list(self, msg: Message) -> None:
         """Handler for received PEP OMEMO_DEVICES_NS payloads"""
@@ -400,7 +449,7 @@ class XEP_0384(BasePlugin):
     async def _read_device_list(self, jid: JID, items: Union[Items, EventItems]) -> None:
         """Read items and devices if we need to set the device list again or not"""
         bare_jid = jid.bare
-        self._store_device_ids(bare_jid, items)
+        await self._store_device_ids(bare_jid, items)
 
         items = list(items)
         device_ids = []
@@ -421,15 +470,15 @@ class XEP_0384(BasePlugin):
                 own_jid.bare, OMEMO_DEVICES_NS,
             )
             items = iq['pubsub']['items']
-            self._store_device_ids(own_jid.bare, items)
+            await self._store_device_ids(own_jid.bare, items)
         except IqError as iq_err:
             if iq_err.condition == "item-not-found":
-                self._store_device_ids(own_jid.bare, [])
+                await self._store_device_ids(own_jid.bare, [])
             else:
                 return  # XXX: Handle this!
 
         if device_ids is None:
-            device_ids = self.get_device_list(own_jid)
+            device_ids = await self.get_device_list(own_jid)
 
         devices = []
         for i in device_ids:
@@ -440,7 +489,7 @@ class XEP_0384(BasePlugin):
         payload['devices'] = devices
 
         jid = self.xmpp.boundjid
-        disco = await self.xmpp['xep_0030'].get_info(jid.bare)
+        disco = await self.xmpp['xep_0030'].get_info(jid=jid.bare, local=False)
         publish_options = PUBLISH_OPTIONS_NODE in disco['disco_info'].get_features()
 
         options = None
@@ -455,6 +504,7 @@ class XEP_0384(BasePlugin):
             })
 
         try:
+            log.debug('Setting own device list to %r', device_ids)
             await self.xmpp['xep_0060'].publish(
                 own_jid.bare, OMEMO_DEVICES_NS, payload=payload, options=options,
             )
@@ -477,17 +527,98 @@ class XEP_0384(BasePlugin):
                     own_jid.bare, OMEMO_DEVICES_NS, payload=payload,
                 )
 
-    def get_device_list(self, jid: JID) -> List[str]:
+    async def get_device_list(self, jid: JID) -> List[str]:
         """Return active device ids. Always contains our own device id."""
-        return self._omemo.getDevices(jid.bare).get('active', [])
+        devices = await self._omemo().getDevices(jid.bare)
+        return devices.get('active', [])
 
-    def trust(self, jid: JID, device_id: int, ik: bytes) -> None:
-        self._omemo.setTrust(jid.bare, device_id, ik, True)
+    async def _should_heartbeat(self, jid: JID, device_id: int, prekey: bool) -> bool:
+        """
+            Internal helper for :py:func:`XEP_0384.should_heartbeat`.
 
-    def distrust(self, jid: JID, device_id: int, ik: bytes) -> None:
-        self._omemo.setTrust(jid.bare, device_id, ik, False)
+            Returns whether we should send a heartbeat message for (JID,
+            device_id).
 
-    def get_trust_for_jid(self, jid: JID) -> Dict[str, List[Optional[Dict[str, Any]]]]:
+            We check if the message is a prekey message, in which case we
+            assume it's a new session and we want to ACK relatively early.
+
+            Otherwise we look at the number of messages since we have last
+            replied and if above a certain threshold we notify them that we're
+            still active.
+        """
+
+        length = await self._omemo().receiving_chain_length(jid.bare, device_id)
+        inactive_session = (length or 0) > self.heartbeat_after
+        log.debug(
+            'Chain length for %r / %d: %d -> inactive_session? %r',
+            jid, device_id, length, inactive_session,
+        )
+        log.debug('Is this a prekey message: %r', prekey)
+
+        res = prekey or inactive_session
+        log.debug('Should heartbeat? %r', res)
+
+        return res
+
+    async def should_heartbeat(self, jid: JID, msg: Union[Message, Encrypted]) -> bool:
+        """
+            Returns whether we should send a heartbeat message to the sender
+            device. See notes about heartbeat in
+            https://xmpp.org/extensions/xep-0384.html#rules.
+
+            This method will return True if this session (to the sender
+            device) is not yet confirmed, or if it hasn't been answered in a
+            while.
+        """
+
+        prekey: bool = False
+
+        # Get prekey information from message
+        encrypted = msg
+        if isinstance(msg, Message):
+            encrypted = msg['omemo_encrypted']
+
+        header = encrypted['header']
+        sid = header['sid']
+        key = header.xml.find("{%s}key[@rid='%s']" % (
+            OMEMO_BASE_NS, self._device_id))
+        # Don't error out. If it's not encrypted to us we don't need to send a
+        # heartbeat.
+        prekey = False
+        if key is not None:
+            key = Key(key)
+            prekey = key['prekey'] in TRUE_VALUES
+
+        return await self._should_heartbeat(jid, sid, prekey)
+
+    async def make_heartbeat(self, jid: JID, device_id: int) -> Message:
+        """
+            Returns a heartbeat message.
+
+            This is mainly used to tell receiving clients that our device is
+            still active. This is an empty key transport message of which we
+            won't use the generated shared secret.
+        """
+
+        msg = self.xmpp.make_message(mto=jid)
+        encrypted = await self.encrypt_message(
+            plaintext=None,
+            recipients=[jid],
+            expect_problems=None,
+            _ignore_trust=True,
+            _device_id=device_id,
+        )
+        msg.append(encrypted)
+        msg.enable('store')
+        return msg
+
+    async def trust(self, jid: JID, device_id: int, ik: bytes) -> None:
+        await self._omemo().setTrust(jid.bare, device_id, ik, True)
+
+    async def distrust(self, jid: JID, device_id: int, ik: bytes) -> None:
+        await self._omemo().setTrust(jid.bare, device_id, ik, False)
+
+    async def get_trust_for_jid(self, jid: JID) -> Dict[str, List[Optional[Dict[str, Any]]]]:
         """
             Fetches trust for JID. The returned dictionary will contain active
             and inactive devices. Each of these dict will contain device ids
@@ -508,31 +639,23 @@ class XEP_0384(BasePlugin):
             }
         """
 
-        return self._omemo.getTrustForJID(jid.bare)
+        return await self._omemo().getTrustForJID(jid.bare)
 
-    def deleteSession(self, bare_jid: JID, device_id: int) -> None:
-        """
-            It might be useful sometimes to be able to reset a broken session.
-            Note that the library won't be able to decrypt any message that is
-            received until a new session is initiated. It is recommended to
-            use `encrypt_ratchet_forwarding_message` to generate a payload to
-            send to this same device you're removing session material for.
-        """
-        pass
-
-    def is_encrypted(self, msg: Message) -> bool:
+    @classmethod
+    def is_encrypted(cls, msg: Message) -> bool:
         return msg.xml.find('{%s}encrypted' % OMEMO_BASE_NS) is not None
 
-    def decrypt_message(
+    async def decrypt_message(
         self,
         encrypted: Encrypted,
         sender: JID,
         allow_untrusted: bool = False,
     ) -> Optional[str]:
         header = encrypted['header']
-        if encrypted['payload']['value'] is None:
-            raise ErroneousPayload('The payload element was empty')
-        payload = b64dec(encrypted['payload']['value'])
+
+        payload = None
+        if encrypted['payload']['value'] is not None:
+            payload = b64dec(encrypted['payload']['value'])
 
         jid = sender.bare
         sid = int(header['sid'])
@@ -554,16 +677,27 @@ class XEP_0384(BasePlugin):
         # XXX: 'cipher' is part of KeyTransportMessages and is used when no payload
         # is passed. We do not implement this yet.
         try:
-            body = self._omemo.decryptMessage(
-                jid,
-                sid,
-                iv,
-                message,
-                isPrekeyMessage,
-                payload,
-                allow_untrusted=allow_untrusted,
-            )
-            return body
+            log.debug('Decryption: Attempt to decrypt message from JID: %r', sender)
+            if payload is None:
+                await self._omemo().decryptRatchetForwardingMessage(
+                    jid,
+                    sid,
+                    iv,
+                    message,
+                    isPrekeyMessage,
+                    allow_untrusted=allow_untrusted,
+                )
+                body = None
+            else:
+                body = await self._omemo().decryptMessage(
+                    jid,
+                    sid,
+                    iv,
+                    message,
+                    isPrekeyMessage,
+                    payload,
+                    allow_untrusted=allow_untrusted,
+                )
         except (omemo.exceptions.NoSessionException,):
             # This might happen when the sender is sending using a session
             # that we don't know about (deleted session storage, etc.). In
@@ -572,18 +706,35 @@ class XEP_0384(BasePlugin):
             raise NoAvailableSession(jid, sid)
         except (omemo.exceptions.TrustException,) as exn:
             if exn.problem == 'undecided':
+                log.debug('Decryption: trust state for JID: %r, device: %r, is undecided', exn.bare_jid, exn.device)
                 raise UndecidedException(exn.bare_jid, exn.device, exn.ik)
             if exn.problem == 'untrusted':
+                log.debug('Decryption: trust state for JID: %r, device: %r, set to untrusted', exn.bare_jid, exn.device)
                 raise UntrustedException(exn.bare_jid, exn.device, exn.ik)
             raise
         finally:
             asyncio.ensure_future(self._publish_bundle())
 
+        if self.auto_heartbeat:
+            log.debug('Checking if heartbeat is required. auto_hearbeat enabled.')
+            should_heartbeat = await self._should_heartbeat(sender, sid, isPrekeyMessage)
+            if should_heartbeat:
+                log.debug('Decryption: Sending hearbeat to %s / %d', jid, sid)
+                async def send_heartbeat():
+                    log.debug('Sending a heartbeat message')
+                    msg = await self.make_heartbeat(JID(jid), sid)
+                    msg.send()
+                asyncio.ensure_future(send_heartbeat())
+
+        return body
+
     async def encrypt_message(
         self,
-        plaintext: str,
+        plaintext: Optional[str],
         recipients: List[JID],
         expect_problems: Optional[Dict[JID, List[int]]] = None,
+        _ignore_trust: bool = False,
+        _device_id: Optional[int] = None,
     ) -> Encrypted:
         """
         Returns an encrypted payload to be placed into a message.
@@ -591,9 +742,22 @@ class XEP_0384(BasePlugin):
         The API for getting an encrypted payload consists of trying first
         and fixing errors progressively. The actual sending happens once the
         application (us) thinks we're good to go.
+
+        If `plaintext` is specified, this will generate a full OMEMO payload. If
+        not, if `_ignore_trust` is True, this will generate a ratchet forwarding
+        message, and otherwise it will generate a key transport message.
+
+        These are rather technical details to the user and fiddling with
+        parameters else than `plaintext` and `recipients` should be rarely
+        needed.
+
+        The `_device_id` parameter is required in the case of a ratchet
+        forwarding message. That is, `plaintext` to None, and `_ignore_trust`
+        to True. If specified, a single recipient JID is required. If not all
+        these conditions are met, ErroneousParameter will be raised.
         """
 
-        recipients = [jid.bare for jid in recipients]
+        barejids: List[str] = [jid.bare for jid in recipients]
 
         old_errors = None  # type: Optional[List[Tuple[Exception, Any, Any]]]
         while True:
@@ -607,29 +771,51 @@ class XEP_0384(BasePlugin):
             expect_problems = {jid.bare: did for (jid, did) in expect_problems.items()}
 
             try:
-                encrypted = self._omemo.encryptMessage(
-                    recipients,
-                    plaintext.encode('utf-8'),
-                    self.bundles,
-                    expect_problems=expect_problems,
-                )
+                log.debug('Encryption: attempt to encrypt for JIDs: %r', barejids)
+                if plaintext is not None:
+                    encrypted = await self._omemo().encryptMessage(
+                        barejids,
+                        plaintext.encode('utf-8'),
+                        bundles=self.bundles,
+                        expect_problems=expect_problems,
+                    )
+                elif _ignore_trust:
+                    if not _device_id or len(barejids) != 1:
+                        raise ErroneousParameter
+                    encrypted = await self._omemo().encryptRatchetForwardingMessage(
+                        bare_jid=barejids[0],
+                        device_id=_device_id,
+                        bundle=None,
+                    )
+                else:
+                    encrypted = await self._omemo().encryptKeyTransportMessage(
+                        barejids,
+                        bundles=self.bundles,
+                        expect_problems=expect_problems,
+                    )
                 return _generate_encrypted_payload(encrypted)
             except omemo.exceptions.EncryptionProblemsException as exception:
                 errors = exception.problems
 
             if errors == old_errors:
+                log.debug('Encryption: Still not possible after another iteration.')
                 raise EncryptionPrepareException(errors)
 
             old_errors = errors
 
             for exn in errors:
                 if isinstance(exn, omemo.exceptions.NoDevicesException):
+                    log.debug('Encryption: Missing device list for JID: %r', exn.bare_jid)
                     await self._fetch_device_list(JID(exn.bare_jid))
                 elif isinstance(exn, omemo.exceptions.MissingBundleException):
+                    log.debug('Encryption: Missing bundle for JID: %r, device: %r', exn.bare_jid, exn.device)
                     bundle = await self._fetch_bundle(exn.bare_jid, exn.device)
                     if bundle is not None:
+                        log.debug('Encryption: Bundle %r found!', exn.device)
                         devices = self.bundles.setdefault(exn.bare_jid, {})
                         devices[exn.device] = bundle
+                    else:
+                        log.debug('Encryption: Bundle %r not found.', exn.device)
                 elif isinstance(exn, omemo.exceptions.TrustException):
                     # On TrustException, there are two possibilities.
                     # Either trust has not been explicitely set yet, and is
@@ -638,6 +824,7 @@ class XEP_0384(BasePlugin):
                     # a choice. If untrusted, then we can safely tell the
                     # OMEMO lib to not encrypt to this device
                     if exn.problem == 'undecided':
+                        log.debug('Encryption: Trust state not set for JID: %r, device: %r', exn.bare_jid, exn.device)
                         raise UndecidedException(exn.bare_jid, exn.device, exn.ik)
                     distrusted_jid = JID(exn.bare_jid)
                     expect_problems.setdefault(distrusted_jid, []).append(exn.device)
